import logging
import os
import unittest
import warnings

from support.logging import get_console_logger
from .commands import ArchiveProject, ScodeProjectUpdater
from .project_fetcher import ArchiveProjectFetcher
from .test_projects import ScienceTestProject, ScodeTestProject, AlmaTestProject
from schema import Author, ArchiveDBSession
from sqlalchemy import exc as sa_exc

_LOG = get_console_logger("scode_project_updater_tests", logging.DEBUG)

class UpdaterTestCase(unittest.TestCase):

    @classmethod
    def setUp(self) -> None:
        self.profile = os.environ['CAPO_PROFILE']

    def restore_scode_project_to_original_state(self):
        self.archive_context = ArchiveDBSession('SDM', profile=self.profile)
        self.pst_context = ArchiveDBSession('PST', profile=self.profile)

        scode_project = ScodeTestProject()
        # start by restoring the title and abstract; authors will need special treatment
        # args = scp.make_args(False)
        args = ['-C', scode_project.project_code, '-P', self.profile, '-T', scode_project.title, '-A', scode_project.abstract]
        project = ScodeProjectUpdater(args=args).update_project()

        with warnings.catch_warnings(), self.archive_context, self.pst_context:
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            # clear existing investigators
            investigators_list = self.archive_context.session.query(Author) \
                .filter(Author.project_code == scode_project.project_code) \
                .all()
            for inv in investigators_list:
                self.archive_context.session.delete(inv)
            self.archive_context.session.commit()

            # insert the canonical ones
            canonical_authors = scode_project.authors
            for author in canonical_authors:
                author.project = project
                self.archive_context.session.add(author)
            self.archive_context.session.commit()

        # confirm restoration
        fetcher = ArchiveProjectFetcher(self.profile)
        restored_project = fetcher.fetch_project(scode_project.project_code)
        self.assertEqual(scode_project.title, restored_project.title)
        self.assertEqual(scode_project.abstract, restored_project.abstract)

        restored_authors = fetcher.detachable_author_list
        self.assertEqual(len(scode_project.authors), len(restored_authors))

        pi_found = False
        for author in restored_authors:
            if author.username == 'srandall':
                self.assertTrue(author.is_pi, 'author is pi')
                self.assertEqual(65409, author.author_id, 'expecting author_id 65409')
                self.assertEqual('4686', author.pst_person_id, 'expecting pst_person_id 4686')
                pi_found = True
            else:
                self.assertFalse(author.is_pi, 'author is pi')
                self.assertNotEqual(65409, author.author_id, 'expecting author_id not 65409')
                self.assertTrue(author.pst_person_id in ('317','341','11991'), "expecting pst_person_ids 317, 341, 11991")

        self.assertTrue(pi_found)

    def restore_science_project_to_original_state(self):
        science_project = ScienceTestProject()
        self.archive_context = ArchiveDBSession('SDM', profile=self.profile)
        self.pst_context = ArchiveDBSession('PST', profile=self.profile)

        # start by restoring the title and abstract; authors will need special treatment
        # args = scp.make_args(False)
        args = ['-C', science_project.project_code, '-P', self.profile, '-T', science_project.title, '-A', science_project.abstract]
        project = ScodeProjectUpdater(args=args).update_project()

        with warnings.catch_warnings(), self.archive_context, self.pst_context:
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)
            # clear existing investigators
            investigators_list = self.archive_context.session.query(Author) \
                .filter(Author.project_code == science_project.project_code) \
                .all()
            for inv in investigators_list:
                self.archive_context.session.delete(inv)
            self.archive_context.session.commit()

            # insert the canonical ones
            canonical_authors = science_project.authors
            for author in canonical_authors:
                author.project = project
                self.archive_context.session.add(author)
            self.archive_context.session.commit()

        # confirm restoration
        fetcher = ArchiveProjectFetcher(self.profile)
        restored_project = fetcher.fetch_project(science_project.project_code)
        self.assertEqual(science_project.title, restored_project.title)
        self.assertEqual(science_project.abstract, restored_project.abstract)
        restored_authors = fetcher.detachable_author_list
        self.assertEqual(len(science_project.authors), len(restored_authors))

        pi_found = False
        for author in restored_authors:
            if author.username == 'jstrader':
                self.assertTrue(author.is_pi, 'author is pi')
                self.assertEqual(8749, author.author_id, 'expecting author_id 8749')
                self.assertEqual('4064', author.pst_person_id, 'expecting pst_person_id 4064')
                pi_found = True
            else:
                self.assertFalse(author.is_pi, 'author is pi')
                self.assertNotEqual(8749, author.pst_person_id, 'expecting author_id not 8749')
                self.assertTrue(8742 < int(author.author_id) < 8751, 'expecting pst_person_id between 8743 and 8750')

        self.assertTrue(pi_found)

    def test_alma_project_has_not_changed(self):
        alma_test_project = AlmaTestProject()
        fetcher = ArchiveProjectFetcher(self.profile)
        fetched = fetcher.fetch_project(alma_test_project.project_code)
        self.assertEqual(alma_test_project.title, fetched.title)
        self.assertEqual(alma_test_project.abstract, fetched.abstract)
        authors_list = fetcher.detachable_author_list
        self.assertEqual(len(alma_test_project.authors), len(authors_list))

        pi_found = False
        for author in authors_list:
            self.assertIsNone(author.pst_person_id, 'expecting no pst_person_id')
            if author.username == 'clada':
                self.assertTrue(author.is_pi, 'author is pi')
                self.assertEqual(37200, author.author_id, 'expecting author_id 37200')
                pi_found = True
            else:
                self.assertFalse(author.is_pi, 'author is pi')
                self.assertNotEqual(37200, author.pst_person_id, 'expecting author_id not 8749')

        self.assertTrue(pi_found)

    def test_restores_scode_project_correctly(self):
        self.restore_scode_project_to_original_state()

    def test_restores_science_project_correctly(self):
        self.restore_science_project_to_original_state()

    def test_can_fetch_from_project_code(self):
        scode_project = ScodeTestProject()
        args = ['-C', scode_project.project_code, '-P', self.profile]
        fetched = ScodeProjectUpdater(args=args).update_project()
        self.assertEqual(scode_project.title, fetched.title)
        self.assertEqual(scode_project.abstract, fetched.abstract)

    def test_can_fetch_non_scode_project(self):
        args = ['-C', '13B-014', '-P', self.profile]
        fetched = ScodeProjectUpdater(args=args).update_project()
        self.assertIsNotNone(fetched)

    def test_can_modify_non_scode_project(self):
        to_modify = ScienceTestProject()
        authors = to_modify.authors.copy()
        pst_ids = [author.pst_person_id for author in authors]
        scp = ArchiveProject(to_modify.project_code, 'foo', to_modify.abstract, pst_ids)
        scp_args = scp.make_args(False)

        updater = ScodeProjectUpdater(args=scp_args)
        updated_project = updater.update_project()

        fetcher = ArchiveProjectFetcher(self.profile)
        retrieved_project = fetcher.fetch_project(updated_project.project_code)
        # title should have changed
        self.assertEqual('foo', retrieved_project.title)

        self.restore_science_project_to_original_state()

    def test_no_update_with_dry_run(self):

        self.restore_scode_project_to_original_state()
        scode_project = ScodeTestProject()
        authors = scode_project.authors.copy()
        pst_ids = [author.pst_person_id for author in authors]
        pst_ids.append(5654)

        scp = ArchiveProject(scode_project.project_code, scode_project.title, scode_project.abstract, pst_ids)
        scp_args = scp.make_args(True)

        updater = ScodeProjectUpdater(args=scp_args)
        updated_project = updater.update_project()

        fetcher = ArchiveProjectFetcher(self.profile)
        retrieved_project = fetcher.fetch_project(updated_project.project_code)
        authors = retrieved_project.authors

        self.assertEqual(len(scode_project.authors), len(authors))
        for author in authors:
            if author.username == 'srandall':
                assert author.is_pi
            else:
                assert not author.is_pi

            self.assertEqual(scode_project.title, retrieved_project.title)
            self.assertEqual(scode_project.abstract, retrieved_project.abstract)

    def test_setting_investigators_preserves_title_and_abstract(self):
        scode_project = ScodeTestProject()
        authors = scode_project.authors.copy()
        pst_ids = [author.pst_person_id for author in authors]
        pst_ids.append(5654)

        scp = ArchiveProject(scode_project.project_code, scode_project.title, scode_project.abstract, pst_ids)
        scp_args = scp.make_args(False)

        updater = ScodeProjectUpdater(args=scp_args)
        updater.update_project()

        fetcher = ArchiveProjectFetcher(self.profile)
        retrieved_project = fetcher.fetch_project(scp.project_code)

        authors = fetcher.detachable_author_list
        self.assertEqual(len(scode_project.authors) + 1, len(authors))
        for author in authors:
            if author.username == 'srandall':
                assert author.is_pi
            else:
                assert not author.is_pi

        self.assertEqual(scode_project.title, retrieved_project.title)
        self.assertEqual(scode_project.abstract, fetcher.abstract)
        self.restore_scode_project_to_original_state()

    def test_alma_project_not_updated(self):
        alma_project = AlmaTestProject()
        args = alma_project.make_args(False)
        try:
            updater = ScodeProjectUpdater(args=args)
            updater.update_project()
        except Exception as exc:
            _LOG.info(f'attempt to update ALMA project failed, as expected: {exc}')
        self.test_alma_project_has_not_changed()

    def test_output_is_desired_format(self):
        scode_project = ScodeTestProject()
        authors = scode_project.authors.copy()
        pi = authors[0]
        coi_pst_ids = [int(coi.pst_person_id) for coi in authors[1:]]
        coi_pst_ids = sorted(coi_pst_ids)

        author_pst_ids = [int(pi.pst_person_id)]
        [author_pst_ids.append(id) for id in coi_pst_ids]
        authors_to_print = [str(id) for id in author_pst_ids]
        id_list = ' '.join(authors_to_print)

        scp = ArchiveProject(scode_project.project_code, scode_project.title, scode_project.abstract, author_pst_ids)
        scp_args = scp.make_args(True)
        ScodeProjectUpdater(args=scp_args).update_project()

        fetcher = ArchiveProjectFetcher(self.profile)
        fetcher.fetch_project(scp.project_code)
        output = fetcher.build_project_info()
        self.assertEqual(3, len(output))
        authors_line = output[2]
        self.assertEqual(f'Authors: {id_list}', authors_line )


UpdaterTestCase()

if __name__ == '__main__':
    unittest.main()
